import javax.microedition.midlet.*;
import javax.microedition.lcdui.*;

public class ArenaDisplay extends Canvas {
	// The font we use to draw with.
	private Font defFont;

	// The size of each tile.
	private int tileWidth, tileHeight; 

	// The number of tiles visible on the screen (rounded up).
	private int tileCountX, tileCountY;

	// The offset of the top-left tile, with respect to the top-left corner
	// of the screen (usually negative).
	private int tileOffsetX, tileOffsetY;

	// How close can the player get to the edges of the level before we
	// stop scrolling.
	private int bufferX, bufferY;

	// The map we're going to draw.
	private Arena arena;
	private int[] palette;

	private Main controller;

	private int tileCount(int screen, int tile) {
		// We want tileCount to always be the smallest odd number
		// greater or equal to the number of tiles that will fit
		// horizontally across the screen.
		int res;

		res = screen/tile;

		// If there's not an integer number of tiles, add a tile.
		if ((screen % tile) != 0) res++;

		// Make sure we have an odd number of tiles.
		if ((res & 1) != 1) res++;

		return res;
	}

	public ArenaDisplay(Arena newarena, int[] newpalette, 
			Main newcontroller) {

		arena = newarena;
		palette = newpalette;
		controller = newcontroller;

		// We want to draw as much as possible in a grid, so we need
		// the smallest font we can get.
		defFont = Font.getFont(
				Font.FACE_MONOSPACE, 
				Font.STYLE_PLAIN, 
				Font.SIZE_SMALL
				);

		tileHeight = defFont.getHeight();

		// Just because we asked for a monospaced font doesn't mean
		// we'll get it. Sadly, checking the width of every possible
		// character takes far too long, so let's just assume that 'm'
		// is the widest character (because it usually is).
		tileWidth = defFont.charWidth('m');

		//System.out.println("Screen size: " + 
		//		getWidth()+"x"+getHeight());
		//System.out.println("Tile size: " + 
		//		tileWidth+"x"+tileHeight);

		// Figure out how many tiles we can squeeze onto the screen,
		// and where. We want the final layout of the display to look
		// like this:
		//
		// ,---------------------.
		// | |     |     |     | |
		// |-+-----+-----+-----+-|
		// | |     |  O/ |     | |
		// | |     | /|  |     | |
		// | |     | / \ |     | |
		// |-+-----+-----+-----+-|
		// | |     |     |     | |
		// `---------------------'
		//
		tileCountX = tileCount(getWidth(), tileWidth);
		tileCountY = tileCount(getHeight(), tileHeight);

		if (tileCountX > arena.getWidth())
			tileCountX = arena.getWidth();
		if (tileCountY > arena.getHeight())
			tileCountY = arena.getHeight();

		//System.out.println("Tiles visible: " + 
		//		tileCountX+"x"+tileCountY);

		tileOffsetX = (getWidth()  - tileCountX*tileWidth)/2 + 
			tileWidth/2; 	// we want to draw at the centre of each
					// tile, not the left edge.
		tileOffsetY = (getHeight() - tileCountY*tileHeight)/2;

		// bufferX/Y will always be integers, since tileCountX/Y are
		// always odd.
		bufferX = (tileCountX-1)/2; 
		bufferY = (tileCountY-1)/2;
	}

	protected void paint(Graphics g) {
		int rX = arena.getRobotX();
		int rY = arena.getRobotY();
		int tileType;
		
		g.setColor(palette[0]);
		g.fillRect(0,0, getWidth(), getHeight());

		// To choose which part of the map we must draw, first we need
		// to know where the player is.

		// The place in the tile grid we start drawing from.
		int tileStartX, tileStartY;

		// The choice of tileStartX/Y is rather involved. If Robot is
		// somewhere in the middle of the level, we want the screen to
		// be perfectly centered on him. However, if he's near the edge
		// of the arena, we want to keep as much of the arena visible
		// as possible. 
		//
		// Note that occasionally tileStartX/Y can be set to a value
		// outside the bounds of the array - this is so that the
		// outermost edges of the arena will be fully displayed, even
		// if the screen is not an integer number of tiles wide.
		if (rX < bufferX - 1) {
			tileStartX = -1;
		} else if (rX > ((arena.getWidth()-1) - bufferX)) {
			tileStartX = (arena.getWidth()-1) - tileCountX + 2;
		} else {
			tileStartX = rX - (tileCountX-1)/2;
		}

		if (rY < bufferY - 1) {
			tileStartY = -1;
		} else if (rY > ((arena.getHeight()-1) - bufferY)) {
			tileStartY = (arena.getHeight()-1) - tileCountY + 2;
		} else {
			tileStartY = rY - (tileCountY-1)/2;
		}

		g.setFont(defFont);

		// Iterate through the chosen chunk of the arena and draw the
		// tiles.
		for (int y = tileStartY; y < tileStartY + tileCountY; y++) {
			if (y < 0) continue;
			if (y > arena.getHeight()-1) break;

			for (int x = tileStartX; x < tileStartX + tileCountX; x++) {
				if (x < 0) continue;
				if (x > arena.getWidth()-1) break;

				switch (arena.typeAt(x,y)) {
					case Arena.TYPE_KITTEN:
					case Arena.TYPE_NKI:
						drawTile(g, arena.glyphAt(x,y), 
							arena.colorAt(x,y), 
							x, y, tileStartX, tileStartY);
						break;

					case Arena.TYPE_WALL:
						drawTile(g, arena.glyphAt(x,y), -1, 
							x, y, tileStartX, tileStartY);
						break;
				}
			}
		}


		drawTile(g, '#', -1, 
				rX, rY, tileStartX, tileStartY);
	}

	private void drawTile(Graphics g, char glyph, int colorIndex, 
			int arenaX, int arenaY, int arenaStartX, int arenaStartY) {

		if (colorIndex == -1) {
			g.setColor(palette[palette.length-1]);
		} else {
			g.setColor(palette[(colorIndex % (palette.length-1))+1]);
		}

		g.drawChar(glyph, 
			(tileOffsetX + (arenaX-arenaStartX)*tileWidth),
			(tileOffsetY + (arenaY-arenaStartY)*tileHeight),
			g.TOP|g.HCENTER);
	}

	protected void keyPressed (int keyCode) {
		int action = getGameAction(keyCode);
		int rX = arena.getRobotX();
		int rY = arena.getRobotY();

		switch(action) {
			case UP:    
				if (rY > 0) rY--; 
				break;
			case DOWN:  
				if (rY < arena.getHeight()-1) rY++;
				break;
			case LEFT:
				if (rX > 0) rX--; 
				break;
			case RIGHT:
				if (rX < arena.getWidth()-1) rX++;
				break;
			default:
				break;
		}

		switch(arena.typeAt(rX, rY)) {
			case Arena.TYPE_ROBOT:
			case Arena.TYPE_WALL:
				return;
			case Arena.TYPE_KITTEN:
				controller.foundKitten();
				break;
			case Arena.TYPE_NKI:
				controller.foundNKI(arena.messageAt(rX, rY));
				break;
			case Arena.TYPE_NOTHING:
				arena.setRobotX(rX);
				arena.setRobotY(rY);
				repaint();
		}
	}

	protected void keyRepeated (int keyCode) {
		keyPressed(keyCode);
	}
}
